# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import platform
from shutil import which
import subprocess

from colcon_core.logging import colcon_logger
from colcon_core.plugin_system import satisfies_version
from colcon_core.plugin_system import SkipExtensionException
from colcon_notification.desktop_notification \
    import DesktopNotificationExtensionPoint

logger = colcon_logger.getChild(__name__)


class NotifySendDesktopNotification(DesktopNotificationExtensionPoint):
    """Use `notify-send` to show notifications."""

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            DesktopNotificationExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')

        if platform.system() != 'Linux':
            raise SkipExtensionException('Not used on non-Linux systems')
        if not which('notify-send'):
            raise SkipExtensionException("Could not find 'notify-send'")

    def notify(self, *, title, message, icon_path=None):  # noqa: D102
        # can't replace a previous notification:
        # https://bugs.launchpad.net/ubuntu/+source/libnotify/+bug/257135
        cmd = [
            'notify-send',
            '-a', 'colcon',
        ]
        if icon_path:
            cmd += ['-i', icon_path]
        cmd += [title, message]

        try:
            subprocess.run(cmd)
        except subprocess.CalledProcessError:
            cmd_str = ' '.join(cmd)
            logger.error(
                "Failed to invoke '{cmd_str}'".format_map(locals()))



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import sys
import time

from colcon_core.logging import colcon_logger
from colcon_core.plugin_system import satisfies_version
from colcon_core.plugin_system import SkipExtensionException
from colcon_notification.desktop_notification \
    import DesktopNotificationExtensionPoint

logger = colcon_logger.getChild(__name__)


class Win32DesktopNotification(DesktopNotificationExtensionPoint):
    """Use the Windows API to show notifications."""

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            DesktopNotificationExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')

        if sys.platform != 'win32':
            raise SkipExtensionException('Not used on non-Windows systems')

        self._initialized = False
        self._last_notification = None

    def notify(self, *, title, message, icon_path=None):  # noqa: D102
        NotificationWindow(title, message, icon_path=icon_path)


class NotificationWindow:
    """Notification windows using the Win32 API."""

    def __init__(self, title, message, icon_path=None):  # noqa: D107
        try:
            import win32con
        except ImportError as e:  # noqa: F841
            logger.debug(
                'Failed to import win32con: {e}'.format_map(locals()))
            return
        try:
            import win32gui
        except ImportError as e:  # noqa: F841
            logger.debug(
                'Failed to import win32gui: {e}'.format_map(locals()))
            return

        wc, class_atom = NotificationWindow._create_window_class()

        # create the window
        hwnd = win32gui.CreateWindow(
            NotificationWindow._class_atom,
            'Colcon notification',  # window class name
            win32con.WS_OVERLAPPED | win32con.WS_SYSMENU,  # style
            0, 0, win32con.CW_USEDEFAULT, win32con.CW_USEDEFAULT,
            0, 0, NotificationWindow._wc.hInstance, None)
        win32gui.UpdateWindow(hwnd)

        # load icon image
        try:
            if icon_path is None:
                # use fall back
                raise Exception()
            hicon = win32gui.LoadImage(
                NotificationWindow._wc.hInstance, icon_path,
                win32con.IMAGE_ICON, 0, 0,
                win32con.LR_LOADFROMFILE | win32con.LR_DEFAULTSIZE,  # flags
            )
        except Exception:  # noqa: B902
            hicon = win32gui.LoadIcon(0, win32con.IDI_APPLICATION)

        # show the notification
        flags = win32gui.NIF_ICON | win32gui.NIF_MESSAGE | win32gui.NIF_TIP
        nid = (hwnd, 0, flags, win32con.WM_USER + 20, hicon, 'tooltip')
        try:
            win32gui.Shell_NotifyIcon(
                win32gui.NIM_ADD, nid)
            win32gui.Shell_NotifyIcon(
                win32gui.NIM_MODIFY, (
                    hwnd, 0, win32gui.NIF_INFO, win32con.WM_USER + 20, hicon,
                    'Balloon  tooltip', message, 200, title))
        except Exception as e:  # noqa: F841
            logger.debug(
                'Failed to show the notification: {e}'.format_map(locals()))
        else:
            # wait a while before destroying the window
            time.sleep(5)
        finally:
            win32gui.DestroyWindow(hwnd)

    _wc = None
    _class_atom = None

    @staticmethod
    def _create_window_class():  # noqa: D102
        import win32api
        import win32con
        import win32gui

        # register the window class once
        if NotificationWindow._wc is None:
            NotificationWindow._wc = win32gui.WNDCLASS()
            NotificationWindow._wc.hInstance = win32api.GetModuleHandle(None)
            NotificationWindow._wc.lpszClassName = 'ColconTaskbar'
            NotificationWindow._wc.lpfnWndProc = {
                win32con.WM_DESTROY: NotificationWindow._on_destroy,
            }
            NotificationWindow._class_atom = win32gui.RegisterClass(
                NotificationWindow._wc)

        return (NotificationWindow._wc, NotificationWindow._class_atom)

    @staticmethod
    def _on_destroy(hwnd, msg, wparam, lparam):  # noqa: D102
        import win32gui
        nid = (hwnd, 0)
        try:
            win32gui.Shell_NotifyIcon(win32gui.NIM_DELETE, nid)
        except Exception as e:  # noqa: F841
            logger.debug(
                'Failed to delete the notification handle: {e}'
                .format_map(locals()))
        win32gui.PostQuitMessage(0)



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from enum import Enum
from pathlib import Path
import traceback

from colcon_core.logging import colcon_logger
from colcon_core.plugin_system import instantiate_extensions
from colcon_core.plugin_system import order_extensions_by_priority

logger = colcon_logger.getChild(__name__)


class Result(Enum):
    """Result type of a job."""

    abort = 1
    failure = 2
    success = 3
    warning = 4

    @staticmethod
    def get_icon_path(result):
        """
        Get the path of the icon file for a specific result.

        :param result: The result enum
        :returns: The path of the icon, or None
        :rtype: str
        """
        if result == Result.abort:
            basename = 'colcon-abort.ico'
        elif result == Result.failure:
            basename = 'colcon-failure.ico'
        elif result == Result.success:
            basename = 'colcon-success.ico'
        elif result == Result.warning:
            basename = 'colcon-warning.ico'
        else:
            return None
        return str(Path(__file__).parent / 'image' / basename)


class DesktopNotificationExtensionPoint:
    """
    The interface for desktop notification extensions.

    A desktop notification extension provides a notification to the user.

    For each instance the attribute `DESKTOP_NOTIFICATION_NAME` is being set
    to the basename of the entry point registering the extension.
    """

    """The version of the desktop notification extension interface."""
    EXTENSION_POINT_VERSION = '1.0'

    """The default priority of desktop notification extensions."""
    PRIORITY = 100

    def notify(self, *, title, message, icon_path=None):
        """
        Create a desktop notification.

        This method must be overridden in a subclass.

        :param str title: The title
        :param str message: The message
        :param str icon_path: The path of the icon
        """
        raise NotImplementedError()


def get_desktop_notification_extensions():
    """
    Get the available desktop notification extensions.

    The extensions are ordered by their priority and entry point name.

    :rtype: OrderedDict
    """
    extensions = instantiate_extensions(__name__)
    for name, extension in extensions.items():
        extension.DESKTOP_NOTIFICATION_NAME = name
    return order_extensions_by_priority(extensions)


def notify(title, message, result):
    """
    Publish a notification message.

    It calls the extensions in order and stops after the first successful
    handling the request.

    :param str title: The title
    :param str message: The message
    :param str icon_path: The path of the icon
    """
    icon_path = Result.get_icon_path(result)
    desktop_notification_extensions = get_desktop_notification_extensions()
    for extension in desktop_notification_extensions.values():
        logger.info(
            'Sending desktop notification using '
            "'{extension.DESKTOP_NOTIFICATION_NAME}'".format_map(locals()))
        try:
            retval = extension.notify(
                title=title, message=message, icon_path=icon_path)
            assert retval is None, 'notify() should return None'
        except Exception as e:  # noqa: F841
            # catch exceptions raised in desktop notification extension
            exc = traceback.format_exc()
            logger.error(
                'Exception in desktop notification extension '
                "'{extension.DESKTOP_NOTIFICATION_NAME}': {e}\n{exc}"
                .format_map(locals()))
            # skip failing extension, continue with next one
        else:
            # only notify using the first extension
            return



# Copyright 2022 Open Source Robotics Foundation, Inc.
# Licensed under the Apache License, Version 2.0

from unittest.mock import call
from unittest.mock import Mock
from unittest.mock import patch

from colcon_core.event.job import JobProgress
from colcon_core.event.job import JobStarted
from colcon_core.event.output import StdoutLine
from colcon_core.event.timer import TimerEvent
from colcon_core.executor import Job
from colcon_notification.event_handler.status import StatusEventHandler


def test_status_event_handler():
    with patch('sys.stdout') as stdout:
        stdout.isatty.return_code = True
        stdout_write = stdout.write

        job = Job(
            identifier='id', dependencies=set(),
            task=Mock(), task_context=Mock())
        job.task.context = job.task_context
        job.task_context.name = job.identifier
        job.task_context.pkg.name = job.identifier

        extension = StatusEventHandler()
        assert extension.enabled

        event = JobStarted(job.identifier)
        extension((event, job))

        event = JobProgress(job.identifier, 'bar')
        extension((event, job))

        event = StdoutLine(b'[  5%] baz')
        extension((event, job))

        event = TimerEvent()
        extension((event, None))
        assert stdout_write.call_args_list[-2:] == [
            call('[0.0s] [0/0 complete] [id:bar 5% - 0.0s]'),
            call('\r'),
        ]
        stdout_write.reset_mock()

        event = StdoutLine(b'[1/10] baz')
        extension((event, job))

        event = TimerEvent()
        extension((event, None))
        assert stdout_write.call_args_list[-2:] == [
            call('[0.0s] [0/0 complete] [id:bar 10% - 0.0s]'),
            call('\r'),
        ]
        stdout_write.reset_mock()

        event = StdoutLine(b'[ 15%] [1/10] ')
        extension((event, job))

        event = TimerEvent()
        extension((event, None))
        assert stdout_write.call_args_list[-2:] == [
            call('[0.0s] [0/0 complete] [id:bar 15% - 0.0s]'),
            call('\r'),
        ]
        stdout_write.reset_mock()



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import os
from pathlib import Path
import platform
import subprocess
import sys

from colcon_core.logging import colcon_logger
from colcon_core.plugin_system import satisfies_version
from colcon_core.plugin_system import SkipExtensionException
import colcon_notification
from colcon_notification.desktop_notification \
    import DesktopNotificationExtensionPoint

logger = colcon_logger.getChild(__name__)


class TerminalNotifierDesktopNotification(DesktopNotificationExtensionPoint):
    """Use `colcon_terminal_notifier` to show notifications."""

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            DesktopNotificationExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')

        if platform.system() != 'Darwin':
            raise SkipExtensionException('Not used on non-Darwin systems')

    def notify(self, *, title, message, icon_path=None):  # noqa: D102
        if title.startswith('-'):
            title = '\\' + title
        if message.startswith('-'):
            message = '\\' + message

        # determine the install prefix of this Python package
        install_prefix = _get_prefix_path(
            Path(colcon_notification.__file__).parent.parent)
        if install_prefix is None:
            # for 'develop' the dist location points to the build directory
            # and the prefix path can't be determined
            # therefore searching the sys.path instead
            for path in sys.path:
                path = _get_prefix_path(path)
                if path is not None and _get_app_path(path):
                    install_prefix = path
                    break

        if install_prefix is None:
            logger.error(
                'Could not determine the install prefix of the '
                'colcon_terminal_notifier.app')
            return

        app_path = _get_app_path(install_prefix)
        if app_path is None:
            logger.error(
                'Could not find the colcon_terminal_notifier.app in the '
                "install prefix '{install_prefix}'".format_map(locals()))
            return

        cmd = [
            'open', str(_get_app_path(install_prefix)),
            '--args',
            '-message', message,
            '-title', title,
            '-group', 'colcon_pid_' + str(os.getpid()),
        ]
        if icon_path:
            cmd += ['--icon', icon_path]

        try:
            subprocess.run(cmd, input=message.encode())
        except FileNotFoundError:
            logger.error(
                "Failed to find 'colcon_terminal_notifier'")
        except subprocess.CalledProcessError:
            cmd_str = ' '.join(cmd)
            logger.error(
                "Failed to invoke '{cmd_str}'".format_map(locals()))


def _get_prefix_path(path):
    path = Path(path)
    if path.parents[1].name.lower() == 'lib':
        return path.parents[2]
    return None


def _get_app_path(prefix_path):
    app_path = prefix_path / 'share' / 'colcon_notification' / \
        'colcon_terminal_notifier.app'
    if app_path.exists():
        return app_path
    return None



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import atexit
import re
import shutil
import sys
import time

from colcon_core.event.job import JobEnded
from colcon_core.event.job import JobProgress
from colcon_core.event.job import JobQueued
from colcon_core.event.job import JobStarted
from colcon_core.event.output import StdoutLine
from colcon_core.event.timer import TimerEvent
from colcon_core.event_handler import EventHandlerExtensionPoint
from colcon_core.event_handler import format_duration
from colcon_core.event_reactor import EventReactorShutdown
from colcon_core.plugin_system import satisfies_version
from colcon_core.subprocess import SIGINT_RESULT


class StatusEventHandler(EventHandlerExtensionPoint):
    """
    Continuously update a status line.

    The extension is only enabled by default if stdout is a tty-like device.

    The extension handles events of the following types:
    - :py:class:`colcon_core.event.output.StdoutLine`
    - :py:class:`colcon_core.event.job.JobEnded`
    - :py:class:`colcon_core.event.job.JobProgress`
    - :py:class:`colcon_core.event.job.JobQueued`
    - :py:class:`colcon_core.event.job.JobStarted`
    - :py:class:`colcon_core.event.timer.TimerEvent`
    - :py:class:`colcon_core.event_reactor.EventReactorShutdown`
    """

    # the priority should be lower than other extensions
    # in order to not update the status before they are finished
    PRIORITY = 50

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            EventHandlerExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')

        self.enabled = sys.stdout.isatty()

        self._start_time = time.monotonic()
        self._queued_count = 0
        self._running = {}
        self._ended = {}

        # pattern to match progress indicator in e.g. make
        self._progress_pattern_percentage = \
            re.compile(r'^\[(  \d| \d\d|1\d\d)%\] ')
        # pattern to match progress indicator in ninja ([X/Y])
        self._progress_pattern_fraction = \
            re.compile(r'^\[([0-9]+)\/([0-9]+)\] ')

        if self.enabled:
            # decorate write methods for stdout / stderr
            # to clear the last status line before other output
            sys.stdout.write = self._write_and_last_clear_status_line(
                sys.stdout.write)
            sys.stdout.buffer.write = self._write_and_last_clear_status_line(
                sys.stdout.buffer.write)
            sys.stderr.write = self._write_and_last_clear_status_line(
                sys.stderr.write)
            sys.stderr.buffer.write = self._write_and_last_clear_status_line(
                sys.stderr.buffer.write)
            self._last_status_line_length = None

            # register exit handle to ensure the last status line is cleared
            atexit.register(self._clear_last_status_line)

    def _write_and_last_clear_status_line(self, func):
        def wrapped_func(*args, **kwargs):
            self._clear_last_status_line()
            return func(*args, **kwargs)
        return wrapped_func

    def _clear_last_status_line(self):
        if self._last_status_line_length is not None:
            # overwrite last status message with spaces
            msg = ' ' * self._last_status_line_length + '\r'
            self._last_status_line_length = None
            sys.stdout.write(msg)

    def __call__(self, event):  # noqa: D102
        data = event[0]

        if isinstance(data, JobQueued):
            self._queued_count += 1

        elif isinstance(data, JobStarted):
            job = event[1]
            assert job not in self._running
            self._running[job] = {'start_time': time.monotonic()}

        elif isinstance(data, JobProgress):
            job = event[1]
            self._running[job]['progress'] = [data.progress]

        elif isinstance(data, StdoutLine):
            line = data.line
            if isinstance(line, bytes):
                line = line.decode(errors='replace')
            match_percentage = self._progress_pattern_percentage.match(line)
            if not match_percentage:
                match_fraction = self._progress_pattern_fraction.match(line)
                if not match_fraction:
                    return
            job = event[1]
            progress = self._running[job].get('progress', [])
            while len(progress) > 1:
                progress.pop()
            if match_percentage:
                progress.append(match_percentage.group(1).lstrip() + '%')
            elif match_fraction:
                progress.append('{:.0%}'.format(
                    int(match_fraction.group(1))/int(match_fraction.group(2))))

        elif isinstance(data, JobEnded):
            job = event[1]
            self._ended[job] = self._running[job]
            self._ended[job]['end_time'] = time.monotonic()
            self._ended[job]['rc'] = data.rc
            del self._running[job]

        elif isinstance(data, TimerEvent):
            now = time.monotonic()
            blocks = []

            # runtime in seconds
            duration_string = format_duration(
                now - self._start_time, fixed_decimal_points=1)
            blocks.append('[{duration_string}]'.format_map(locals()))

            # number of completed jobs / number of jobs
            blocks.append(
                '[%d/%d complete]' %
                (len(self._ended), self._queued_count))

            # number of failed jobs if not zero
            failed_jobs = [
                j for j, d in self._ended.items()
                if d['rc'] and d['rc'] != SIGINT_RESULT]
            if failed_jobs:
                blocks.append('[%d failed]' % len(failed_jobs))

            # number of ongoing jobs if greater one
            if len(self._running) > 1:
                blocks.append('[%d ongoing]' % len(self._running))

            # job identifier, label and time for ongoing jobs
            for job, d in self._running.items():
                msg = job.task.context.pkg.name
                if 'progress' in d:
                    msg += ':%s' % ' '.join(d['progress'])
                duration_string = format_duration(
                    now - d['start_time'], fixed_decimal_points=1)
                blocks.append(
                    '[{msg} - {duration_string}]'.format_map(locals()))

            # determine blocks which fit into terminal width
            max_width = shutil.get_terminal_size().columns
            for i in reversed(range(len(blocks))):
                msg = ' '.join(blocks[0:i + 1])
                # append dots when skipping at least one block
                if i < len(blocks) - 1:
                    msg += ' ...'
                if len(msg) < max_width:
                    break
            else:
                return

            print(msg, end='\r')
            self._last_status_line_length = len(msg)

        elif isinstance(data, EventReactorShutdown):
            self._clear_last_status_line()
            atexit.unregister(self._clear_last_status_line)



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

from colcon_core.logging import colcon_logger
from colcon_core.plugin_system import satisfies_version
from colcon_core.plugin_system import SkipExtensionException
from colcon_notification.desktop_notification \
    import DesktopNotificationExtensionPoint

logger = colcon_logger.getChild(__name__)


class Notify2DesktopNotification(DesktopNotificationExtensionPoint):
    """Use `notify2` to show notifications."""

    # the priority is higher then 'notify_send' extension
    PRIORITY = 200

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            DesktopNotificationExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')

        try:
            import notify2  # noqa: F401
        except ImportError:
            raise SkipExtensionException("'notify2' not found")

        self._initialized = False
        self._last_notification = None

    def notify(self, *, title, message, icon_path=None):  # noqa: D102
        import notify2

        if not self._initialized:
            try:
                notify2.init('colcon')
            except Exception as e:  # noqa: F841
                logger.debug(
                    'Failed to initialize notify2: {e}'.format_map(locals()))
                return
            self._initialized = True

        if self._last_notification is None:
            self._last_notification = notify2.Notification(
                title, message=message, icon=icon_path)
            self._last_notification.set_hint_byte('transient', b'\x01')
        else:
            self._last_notification.update(
                title, message=message, icon=icon_path)
        self._last_notification.show()



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import os

from colcon_core.event.job import JobEnded
from colcon_core.event.output import StderrLine
from colcon_core.event.test import TestFailure
from colcon_core.event_handler import EventHandlerExtensionPoint
from colcon_core.event_reactor import EventReactorShutdown
from colcon_core.plugin_system import satisfies_version
from colcon_core.subprocess import SIGINT_RESULT
from colcon_notification.desktop_notification import notify
from colcon_notification.desktop_notification import Result


class DesktopNotificationEventHandler(EventHandlerExtensionPoint):
    """
    Desktop notification of the summary.

    The extension handles events of the following types:
    - :py:class:`colcon_core.event.output.StderrLine`
    - :py:class:`colcon_core.event.job.JobEnded`
    - :py:class:`colcon_core.event.test.TestFailure`
    """

    # the priority should be lower than all status and notification extensions
    # in order to not trigger the notification before they are finished
    PRIORITY = 20

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            EventHandlerExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')
        self._any_stderr_output = False
        self._any_test_failure = False
        self._failed = []
        self._all_failed_aborted = True

    def __call__(self, event):  # noqa: D102
        data = event[0]

        if isinstance(data, StderrLine):
            self._any_stderr_output = True

        elif isinstance(data, TestFailure):
            self._any_test_failure = True

        elif isinstance(data, JobEnded):
            if data.rc:
                job = event[1]
                self._failed.append(job)
                if data.rc != SIGINT_RESULT:
                    self._all_failed_aborted = False

        elif isinstance(data, EventReactorShutdown):
            if self._failed:
                pkg_info = self._failed[0].task.context.pkg.name
                if len(self._failed) > 1:
                    pkg_info += ' + ' + str(len(self._failed) - 1)
                if self._all_failed_aborted:
                    label = 'aborted'
                    result = Result.abort
                else:
                    label = 'failed'
                    result = Result.failure
                notify(
                    '`{context.command_name} {context.args.verb_name}` '
                    '{label} ({pkg_info})'.format(
                        label=label, pkg_info=pkg_info, **self.__dict__),
                    os.getcwd(), result)
            elif not self._any_stderr_output and not self._any_test_failure:
                notify(
                    '`{context.command_name} {context.args.verb_name}` '
                    'successful'.format_map(self.__dict__),
                    os.getcwd(), Result.success)
            else:
                msg_suffix = []
                if self._any_stderr_output:
                    msg_suffix.append('stderr output')
                if self._any_test_failure:
                    msg_suffix.append('test failures')
                msg_suffix = ' and '.join(msg_suffix)
                notify(
                    '`{context.command_name} {context.args.verb_name}` '
                    .format_map(self.__dict__) +
                    'with {msg_suffix}'.format_map(locals()),
                    os.getcwd(), Result.warning)



# Copyright 2016-2018 Dirk Thomas
# Licensed under the Apache License, Version 2.0

import os
import platform
import sys

from colcon_core.event.job import JobEnded
from colcon_core.event.job import JobQueued
from colcon_core.event.job import JobStarted
from colcon_core.event_handler import EventHandlerExtensionPoint
from colcon_core.plugin_system import satisfies_version


class TerminalTitleEventHandler(EventHandlerExtensionPoint):
    """
    Show status in the terminal title.

    The extension is only enabled by default if stdout is a tty-like device
    and not a dumb terminal.

    The extension handles events of the following types:
    - :py:class:`colcon_core.event.job.JobQueued`
    - :py:class:`colcon_core.event.job.JobStarted`
    - :py:class:`colcon_core.event.job.JobEnded`
    """

    # the priority should be lower than all status and notification extensions
    # in order to not trigger the notification before they are finished
    PRIORITY = 20

    def __init__(self):  # noqa: D107
        super().__init__()
        satisfies_version(
            EventHandlerExtensionPoint.EXTENSION_POINT_VERSION, '^1.0')

        # Setting the title does not work in dumb terminals, i.e., Emacs
        self.enabled = sys.stdout.isatty() and os.getenv('TERM') != 'dumb'

        self._queued_count = 0
        self._ongoing_count = 0
        self._ended_count = 0

    def __call__(self, event):  # noqa: D102
        data = event[0]

        if isinstance(data, JobQueued):
            self._queued_count += 1

        elif isinstance(data, JobStarted):
            self._ongoing_count += 1
            self._update()

        elif isinstance(data, JobEnded):
            assert self._ongoing_count > 0
            self._ongoing_count -= 1
            self._ended_count += 1
            self._update()

    def _update(self):
        message = '{context.command_name} {context.args.verb_name} ' \
            '[{_ended_count}/{_queued_count} done] ' \
            '[{_ongoing_count} ongoing]'.format_map(self.__dict__)

        if platform.system() == 'Windows':
            return _windows_update_title(message)
        else:
            return _ascii_update_title(message)


def _ascii_update_title(message):
    sys.stdout.write('\x1b]0;{message}\x07'.format_map(locals()))


def _windows_update_title(message):
    import ctypes
    ctypes.windll.kernel32.SetConsoleTitleW(message)
